# SPDX-License-Identifier: GPL-3.0-or-later
"""
Tests for miscellaneous utility methods.
"""

import pytest
from freedommaker import utils


def test_add_disk_offsets():
    """Test disk offset addition."""
    assert utils.add_disk_offsets('1MiB', '2MiB') == '3MiB'
    # mib is not allowed as unit. Only MiB.
    with pytest.raises(NotImplementedError):
        utils.add_disk_offsets('1mib', '1mib')

    with pytest.raises(NotImplementedError):
        utils.add_disk_offsets('1GiB', '1MiB')

    with pytest.raises(NotImplementedError):
        utils.add_disk_offsets('1MiB', '1GiB')

    with pytest.raises(ValueError):
        utils.add_disk_offsets('xMiB', '1MiB')

    with pytest.raises(ValueError):
        utils.add_disk_offsets('xMiB', 'xMiB')


def test_parse_disk_size():
    """Test parsing disk sizes."""
    assert utils.parse_disk_size(0) == 0
    assert utils.parse_disk_size(123) == 123
    assert utils.parse_disk_size('123') == 123
    assert utils.parse_disk_size('123b') == 123
    assert utils.parse_disk_size('123k') == 123 * 1024
    assert utils.parse_disk_size('123K') == 123 * 1024
    assert utils.parse_disk_size('123M') == 123 * 1024 * 1024
    assert utils.parse_disk_size('123G') == 123 * 1024 * 1024 * 1024
    assert utils.parse_disk_size('123T') == 123 * 1024 * 1024 * 1024 * 1024
    with pytest.raises(ValueError):
        utils.parse_disk_size('-1M')

    with pytest.raises(ValueError):
        utils.parse_disk_size('1B')

    with pytest.raises(ValueError):
        utils.parse_disk_size('1P')

    with pytest.raises(ValueError):
        utils.parse_disk_size('test')


def test_format_disk_size():
    """Test that formatting disk sizes works."""
    assert utils.format_disk_size(0) == '0'
    assert utils.format_disk_size(1) == '1'
    assert utils.format_disk_size(1023) == '1023'
    assert utils.format_disk_size(1024) == '1K'
    assert utils.format_disk_size(1023 * 1023) == '1046529'
    assert utils.format_disk_size(1023 * 1024) == '1023K'
    assert utils.format_disk_size(1024 * 1024) == '1M'
    assert utils.format_disk_size(1023 * 1024 * 1024) == '1023M'
    assert utils.format_disk_size(1024 * 1024 * 1024) == '1G'
    assert utils.format_disk_size(1023 * 1024 * 1024 * 1024) == '1023G'
    assert utils.format_disk_size(1024 * 1024 * 1024 * 1024) == '1T'
    assert utils.format_disk_size(1023 * 1024 * 1024 * 1024 * 1024) == '1023T'


def test_add_disk_sizes():
    """Test that adding two disk sizes works as expected."""
    assert utils.add_disk_sizes('1M', '1K') == '1025K'
    assert utils.add_disk_sizes('1G', '1K') == '1048577K'
    assert utils.add_disk_sizes('512M', '512M') == '1G'
    assert utils.add_disk_sizes('3800M', '1000M') == '4800M'



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Worker class to build Cubietech Cubietruck image.
"""

from .a20 import A20ImageBuilder


class CubietruckImageBuilder(A20ImageBuilder):
    """Image builder for Cubietruck (Cubieboard 3) target."""
    machine = 'cubietruck'
    flash_kernel_name = 'Cubietech Cubietruck'
    u_boot_path = 'usr/lib/u-boot/Cubietruck/u-boot-sunxi-with-spl.bin'



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Worker class to build VirtualBox amd64 images.
"""

from .virtualbox import VirtualBoxImageBuilder


class VirtualBoxAmd64ImageBuilder(VirtualBoxImageBuilder):
    """Image builder for all VirtualBox amd64 targets."""
    architecture = 'amd64'
    kernel_flavor = 'amd64'



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Base worker class to build Raspberry Pi 2 and 3 images.
"""

from .. import library
from .arm import ARMImageBuilder


class RaspberryPiWithUBoot(ARMImageBuilder):
    """Base image builder for Raspberry Pi 2 and 3 targets."""
    include_non_free_firmware = True
    uboot_variant = None
    firmware_filesystem_type = 'vfat'
    firmware_size = '256MiB'
    u_boot_variant = 'rpi'

    def install_boot_loader(self, state):
        """Install the boot loader onto the image."""
        if not self.u_boot_rpi_variant:
            raise NotImplementedError

        firmware_package = 'raspi-firmware'
        script = '''
set -e
set -x
set -o pipefail

apt-get install --no-install-recommends -y dpkg-dev
cd /tmp
apt-get source {firmware_package}
cp {firmware_package}*/boot/* /boot/firmware
rm -rf {firmware_package}*
cd /

# remove unneeded firmware files
rm -f /boot/firmware/fixup_*
rm -f /boot/firmware/start_*

# u-boot setup
apt-get install -y u-boot-rpi
cp /usr/lib/u-boot/{u_boot_rpi_variant}/u-boot.bin /boot/firmware/kernel.img
cp /usr/lib/u-boot/{u_boot_rpi_variant}/u-boot.bin /boot/firmware/kernel7.img
'''.format(firmware_package=firmware_package,
           u_boot_rpi_variant=self.u_boot_rpi_variant)
        library.run_in_chroot(state, ['bash', '-c', script])



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Base worker class to build all ARM images.
"""

from ..builder import ImageBuilder


class ARMImageBuilder(ImageBuilder):
    """Base image builder for all ARM targets."""
    boot_loader = 'u-boot'
    boot_filesystem_type = 'ext2'
    boot_size = '256MiB'

    @classmethod
    def get_target_name(cls):
        """Return the name of the target for an image builder."""
        return getattr(cls, 'machine', None)



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Worker class to build Cubietech Cubieboard2 image.
"""

from .a20 import A20ImageBuilder


class Cubieboard2ImageBuilder(A20ImageBuilder):
    """Image builder for Cubieboard 2 target."""
    machine = 'cubieboard2'
    flash_kernel_name = 'Cubietech Cubieboard2'
    u_boot_path = 'usr/lib/u-boot/Cubieboard2/u-boot-sunxi-with-spl.bin'



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Worker class to build Olimex A20-OLinuXino-LIME image.
"""

from .a20 import A20ImageBuilder


class A20OLinuXinoLimeImageBuilder(A20ImageBuilder):
    """Image builder for A20 OLinuXino Lime targets."""
    machine = 'a20-olinuxino-lime'
    flash_kernel_name = 'Olimex A20-OLinuXino-LIME'
    u_boot_path = 'usr/lib/u-boot/A20-OLinuXino-Lime/u-boot-sunxi-with-spl.bin'



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Basic image builder using internal implementation.
"""

import logging

from . import library, releases, utils

logger = logging.getLogger(__name__)


class InternalBuilderBackend:
    """Build an image using internal implementation."""

    def __init__(self, builder):
        """Initialize the builder."""
        self.builder = builder
        self.state = {'success': True}

    def make_image(self):
        """Create a disk image."""
        # enable systemd resolved?
        try:
            self._get_temp_image_file()
            self._create_empty_image()
            self._create_partitions()
            self._loopback_setup()
            self._create_filesystems()
            self._mount_filesystems()
            self._setup_extra_storage()
            self._debootstrap()
            self._set_hostname()
            self._lock_root_user()
            self._create_sudo_user()
            self._set_freedombox_disk_image_flag()
            self._create_fstab()
            self._mount_additional_filesystems()
            self._setup_build_apt()
            self._install_freedombox_packages()
            if self.builder.arguments.with_build_dep:
                self._install_build_dependencies()
            self._remove_ssh_keys()
            self._install_boot_loader()
            self._extra_setup()
            self._setup_final_apt()
            self._fill_free_space_with_zeros()
        except (Exception, KeyboardInterrupt) as exception:
            logger.exception('Exception during build - %s', exception)
            self.state['success'] = False
            raise
        finally:
            self._teardown()

    def _get_temp_image_file(self):
        """Get the temporary path to where the image should be built.

        If building to RAM is enabled, create a temporary directory, mount
        tmpfs in it and return a path in that directory. This is so that builds
        that happen in RAM will be faster.

        If building to RAM is disabled, append .temp to the final file name and
        return it.

        """
        if not self.builder.arguments.build_in_ram:
            return library.create_temp_image(self.state,
                                             self.builder.image_file)

        size = utils.add_disk_sizes(self.builder.image_size,
                                    self.builder.extra_storage_size)
        size = utils.add_disk_sizes(size, '100M')  # Buffer
        return library.create_ram_directory_image(self.state,
                                                  self.builder.image_file,
                                                  size)

    def _create_empty_image(self):
        """Create an empty disk image to create partitions in."""
        library.create_image(self.state, self.builder.image_size)

    def _create_partitions(self):
        """Create partition table and partitions in the image."""
        # Don't install MBR on the image file, it is not needed as we use
        # either grub, u-boot or UEFI.
        library.create_partition_table(self.state,
                                       self.builder.partition_table_type)
        boot_partition_number = 1

        offset = '4MiB'
        if self.builder.efi_filesystem_type:
            end = utils.add_disk_offsets(offset, self.builder.efi_size)
            library.create_partition(self.state, 'efi', offset, end,
                                     self.builder.efi_filesystem_type)
            offset = end

        if self.builder.firmware_filesystem_type:
            end = utils.add_disk_offsets(offset, self.builder.firmware_size)
            library.create_partition(self.state, 'firmware', offset, end,
                                     self.builder.firmware_filesystem_type)
            offset = end
            boot_partition_number += 1

        if self.builder.boot_filesystem_type:
            end = utils.add_disk_offsets(offset, self.builder.boot_size)
            library.create_partition(self.state, 'boot', offset, end,
                                     self.builder.boot_filesystem_type)
            offset = end

        library.create_partition(self.state, 'root', offset, '100%',
                                 self.builder.root_filesystem_type)

        library.set_boot_flag(self.state,
                              partition_number=boot_partition_number)

    def _loopback_setup(self):
        """Perform mapping to loopback devices from partitions in image file."""
        library.loopback_setup(self.state)

    def _create_filesystems(self):
        """Create file systems inside the partitions created."""
        if self.builder.firmware_filesystem_type:
            library.create_filesystem(self.state['devices']['firmware'],
                                      self.builder.firmware_filesystem_type)

        if self.builder.efi_filesystem_type:
            library.create_filesystem(self.state['devices']['efi'],
                                      self.builder.efi_filesystem_type)

        if self.builder.boot_filesystem_type:
            library.create_filesystem(self.state['devices']['boot'],
                                      self.builder.boot_filesystem_type)

        library.create_filesystem(self.state['devices']['root'],
                                  self.builder.root_filesystem_type)

    def _mount_filesystems(self):
        """Mount the filesystems in the right places."""
        library.mount_filesystem(self.state, 'root', None)

        if self.builder.boot_filesystem_type:
            library.mount_filesystem(self.state, 'boot', 'boot')

        if self.builder.efi_filesystem_type:
            library.mount_filesystem(self.state, 'efi', 'boot/efi')

        if self.builder.firmware_filesystem_type:
            library.mount_filesystem(self.state, 'firmware', 'boot/firmware')

    def _setup_extra_storage(self):
        """Setup some extra storage for root filesystem.

        btrfs runs out of space even when the data stored is not even half of
        the available space. This is due to it needing re-balance operation. It
        may not be possible to run the re-balance during a heavy operation such
        as installing a package with lot of dependencies (freedombox). Adding
        an additional disk is a good work around for that. The additional
        device will be removed during cleanup and filesystem will be
        re-balanced.

        """
        library.setup_extra_storage(self.state,
                                    self.builder.root_filesystem_type,
                                    self.builder.extra_storage_size)

    def _mount_additional_filesystems(self):
        """Mount extra filesystems: dev, devpts, sys and proc."""
        library.mount_filesystem(self.state, '/dev', 'dev', is_bind_mount=True)
        library.mount_filesystem(self.state,
                                 '/dev/pts',
                                 'dev/pts',
                                 is_bind_mount=True)
        library.mount_filesystem(self.state,
                                 '/proc',
                                 'proc',
                                 is_bind_mount=True)
        library.mount_filesystem(self.state, '/sys', 'sys', is_bind_mount=True)

        # Kill all the processes on the / filesystem before attempting to
        # unmount /dev/pts. Otherwise, unmounting /dev/pts will fail.
        library.schedule_cleanup(self.state, library.process_cleanup,
                                 self.state)

    def _get_packages(self):
        """Return the list of extra packages to install.

        XXX: Move this to builder.py.
        """
        return self._get_basic_packages() + \
            self._get_kernel_packages() + \
            self._get_boot_loader_packages() + \
            self._get_filesystem_packages() + \
            self._get_extra_packages()

    def _get_basic_packages(self):
        """Return a list of basic packages for all kinds of images."""
        return self.builder.packages

    def _get_kernel_packages(self):
        """Return package needed for kernel."""
        if not self.builder.kernel_flavor:
            return []

        return ['linux-image-' + self.builder.kernel_flavor]

    def _get_boot_loader_packages(self):
        """Return packages needed by boot loader."""
        if not self.builder.boot_loader:
            return []

        if self.builder.boot_loader == 'grub':
            return ['grub-pc']

        if self.builder.boot_loader == 'u-boot':
            if self.builder.architecture == 'arm64':
                return ['u-boot-sunxi', 'u-boot-tools']

            if self.builder.u_boot_variant == 'omap':
                return ['u-boot-omap', 'u-boot-tools']

            if self.builder.u_boot_variant == 'rpi':
                return ['u-boot-rpi', 'u-boot-tools']

            if self.builder.u_boot_variant == 'sunxi':
                return ['u-boot-sunxi', 'u-boot-tools']

            raise ValueError('U-boot variant must be specified.')

        raise NotImplementedError

    def _get_filesystem_packages(self):
        """Return packages required for filesystems."""
        packages = []
        if 'btrfs' in [
                self.builder.root_filesystem_type,
                self.builder.boot_filesystem_type,
                self.builder.efi_filesystem_type,
                self.builder.firmware_filesystem_type
        ]:
            packages += ['btrfs-progs']

        return packages

    def _get_extra_packages(self):
        """Return extra packages passed on the command line."""
        return self.builder.arguments.package or []

    def _debootstrap(self):
        """Run debootstrap on the mount point."""
        variant = self.builder.debootstrap_variant or '-'
        library.debootstrap(self.state, self.builder.architecture,
                            self.builder.arguments.distribution, variant,
                            self.builder.release_components,
                            self._get_packages(),
                            self.builder.arguments.build_mirror)

    def _set_hostname(self):
        """Set hostname in debootstrapped file system."""
        library.set_hostname(self.state, self.builder.arguments.hostname)

    def _should_use_backports(self):
        """Return whether backports is in use."""
        is_stable = releases.is_stable(self.builder.arguments.distribution)
        return is_stable and not self.builder.arguments.disable_backports

    def _install_freedombox_packages(self):
        """Install freedombox packages.

        Based on the options provided during the build, packages can be from
        the distribution, from backports or supplied as files during the build.
        """
        custom_freedombox_packages = [
            package  # freedombox and freedombox-doc-* deb packages
            for package in (self.builder.arguments.custom_package or [])
            if 'freedombox' in package
        ]
        if custom_freedombox_packages:
            for custom_freedombox_package in custom_freedombox_packages:
                library.install_custom_package(self.state,
                                               custom_freedombox_package)
        else:
            if self._should_use_backports():
                packages = ('freedombox', 'freedombox-doc-en',
                            'freedombox-doc-es')
                library.install_from_backports(self.state, packages)
            else:
                library.install_package(self.state, 'freedombox')

    def _install_build_dependencies(self):
        library.install_package(self.state, 'git')
        library.run_in_chroot(self.state, [
            'git', 'clone', '--depth=1',
            'https://salsa.debian.org/freedombox-team/freedombox.git',
            '/tmp/freedombox'
        ])

        library.run_in_chroot(self.state, [
            'apt-get', 'build-dep', '--no-install-recommends', '--yes',
            '/tmp/freedombox/'
        ])

        # In case new dependencies conflict with old dependencies
        library.run_in_chroot(self.state, ['apt-mark', 'hold', 'freedombox'])
        script = '''apt-get install --no-upgrade --yes \
  $(/tmp/freedombox/run --develop --list-dependencies)'''
        library.run_in_chroot(self.state, ['bash', '-c', script])
        library.run_in_chroot(self.state, ['apt-mark', 'unhold', 'freedombox'])

        library.install_package(self.state, 'ncurses-term')

        # clean up
        library.run_in_chroot(self.state, ['rm', '-rf', '/tmp/freedombox'])

    def _lock_root_user(self):
        """Lock the root user account."""
        logger.info('Locking root user')
        library.run_in_chroot(self.state, ['passwd', '-l', 'root'])

    def _create_sudo_user(self):
        """Create a user in the image with sudo permissions."""
        username = 'fbx'
        logger.info('Creating a new sudo user %s', username)
        library.run_in_chroot(
            self.state,
            ['adduser', '--gecos', username, '--disabled-password', username])
        library.run_in_chroot(self.state, ['adduser', username, 'sudo'])

    def _set_freedombox_disk_image_flag(self):
        """Set a flag to indicate that this is a FreedomBox image.

        And that FreedomBox is not installed using a Debian package.

        """
        library.run_in_chroot(
            self.state, ['mkdir', '-p', '-m', '755', '/var/lib/freedombox'])
        library.run_in_chroot(
            self.state,
            ['touch', '/var/lib/freedombox/is-freedombox-disk-image'])

    def _remove_ssh_keys(self):
        """Remove SSH keys so that images don't contain known keys."""
        library.run_in_chroot(self.state,
                              ['sh', '-c', 'rm -f /etc/ssh/ssh_host_*'],
                              check=False)

    def _create_fstab(self):
        """Create fstab with entries for each partition."""
        library.add_fstab_entry(self.state,
                                'root',
                                self.builder.root_filesystem_type,
                                1,
                                append=False)
        if self.builder.boot_filesystem_type:
            library.add_fstab_entry(self.state, 'boot',
                                    self.builder.boot_filesystem_type, 2)

        if self.builder.efi_filesystem_type:
            library.add_fstab_entry(self.state, 'efi',
                                    self.builder.efi_filesystem_type, 2)

        if self.builder.firmware_filesystem_type:
            library.add_fstab_entry(self.state, 'firmware',
                                    self.builder.firmware_filesystem_type, 2)

    def _install_boot_loader(self):
        """Install grub/u-boot boot loader."""
        if self.builder.boot_loader == 'grub':
            library.install_grub(self.state)

        if hasattr(self.builder, 'flash_kernel_name') and \
           self.builder.flash_kernel_name:
            options = getattr(self.builder, 'flash_kernel_options', None)
            library.setup_flash_kernel(self.state,
                                       self.builder.flash_kernel_name, options,
                                       self.builder.boot_filesystem_type)

        if hasattr(self.builder, 'install_boot_loader'):
            self.builder.install_boot_loader(self.state)

        if hasattr(self.builder, 'update_initramfs') and \
           not self.builder.update_initramfs:
            return

        library.update_initramfs(self.state)

    def _extra_setup(self):
        """Allow builders to run additional setup if needed."""
        if hasattr(self.builder, 'extra_setup'):
            self.builder.extra_setup(self.state)

    def _setup_build_apt(self):
        """Setup apt to use the build mirror."""
        use_backports = self._should_use_backports()
        library.setup_apt(self.state,
                          self.builder.arguments.build_mirror,
                          self.builder.arguments.distribution,
                          self.builder.release_components,
                          enable_backports=use_backports)

    def _setup_final_apt(self):
        """Setup apt to use the image mirror."""
        library.setup_apt(self.state, self.builder.arguments.mirror,
                          self.builder.arguments.distribution,
                          self.builder.release_components)

    def _fill_free_space_with_zeros(self):
        """Fill up the free space in the image with zeros.

        So that we can compress the image better.
        """
        library.fill_free_space_with_zeros(self.state)

    def _teardown(self):
        """Run cleanup operations for each step that executed."""
        library.cleanup(self.state)



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Base worker class to build AMD64/Intel images.
"""

from ..builder import ImageBuilder


class AMDIntelImageBuilder(ImageBuilder):
    """Base image build for all Intel/AMD targets."""
    boot_loader = 'grub'

    @classmethod
    def get_target_name(cls):
        """Return the name of the target for an image builder."""
        return getattr(cls, 'architecture', None)



#!/usr/bin/python3
# SPDX-License-Identifier: GPL-3.0-or-later
"""
Utility for setting passwords inside disk/VM images.

Written for FreedomBox images.  Works on Qemu, VirtualBox and raw disk
images.
"""

import argparse
import getpass
import logging
import os
import subprocess
import sys
import tempfile

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def main():
    """The main entry point."""
    logging.basicConfig(level=logging.INFO)

    parser = argparse.ArgumentParser(
        description='Change password of a user inside a disk image file')
    parser.add_argument(
        'image',
        help='Disk image file (.img or .vdi) inside which user manipulation '
        'is sought')
    parser.add_argument('user', help='User account to change password for')
    parser.add_argument('--password', help='New password for user')

    arguments = parser.parse_args()

    image_type = get_image_type(arguments)

    check_requirements(image_type)

    try:
        password = arguments.password or take_password()
        perform_operations(arguments, password, image_type)
    except subprocess.CalledProcessError as exception:
        logger.error('Error running command: %s', exception.cmd)
        if exception.output:
            logger.error('Error output - %s', exception.output.decode())
        sys.exit(1)
    except KeyboardInterrupt:
        logger.error('Command terminated by user action')
        sys.exit(2)
    except Exception as exception:  # pylint: disable=broad-except
        logger.exception('Unexpected error: %s', exception)
        sys.exit(3)


def check_requirements(image_type):
    """Check that the necessary requirements are available."""
    logger.info('Checking for necessary dependencies')
    if os.geteuid() != 0:
        logger.error('Due to limitations of the tools involved, you need to '
                     'run this command as "root" user or using the "sudo" '
                     'command.')
        sys.exit(-1)

    if image_type == 'raw':
        try:
            subprocess.check_output(['which', 'kpartx'])
        except subprocess.CalledProcessError:
            logger.error('"kpartx" command not found.  On Debian based '
                         'systems it is provided by the package "kpartx".')
            sys.exit(-1)

    if image_type == 'vm':
        try:
            subprocess.check_output(['which', 'qemu-nbd'])
        except subprocess.CalledProcessError:
            logger.error('"qemu-nbd" command not found.  On Debian based '
                         'systems it is provided by the package "qemu-utils".')
            sys.exit(-1)

    found = False
    with open('/proc/filesystems') as file_handle:
        for line in file_handle:
            if 'btrfs' in line.split():
                found = True

    if not found:
        logger.error('Your kernel does not support Btrfs filesystem.')
        sys.exit(-1)


def get_image_type(arguments):
    """Return the type of the disk image: raw/vm."""
    if arguments.image.split('.')[-1] in ('vdi', 'qcow2'):
        return 'vm'

    return 'raw'


def take_password():
    """Prompt for new password to be set."""
    while True:
        password1 = getpass.getpass('Enter new password: ')
        password2 = getpass.getpass('Re-enter new password: ')
        if password1 == password2:
            return password1

        logger.error('Passwords do not match\n')


def perform_operations(arguments, password, image_type):
    """Map/mount image and change password."""
    map_info = map_disk_image(arguments.image, image_type)

    logger.info('Root device is - %s', map_info['root_device'])

    try:
        mount_info = mount_disk_image(map_info['root_device'])

        try:
            change_password(mount_info['root_path'], arguments.user, password)
        finally:
            unmount_disk_image(mount_info)
    finally:
        unmap_disk_image(map_info)


def map_disk_image(disk_image, image_type):
    """Map the partitions inside disk image as block devices."""
    if image_type == 'vm':
        return map_vm_disk_image(disk_image)

    return map_raw_disk_image(disk_image)


def map_vm_disk_image(disk_image):
    """Map the partitions inside a VM disk image as block devices."""
    logger.info('Adding partition mappings for VM disk image - %s', disk_image)
    device = '/dev/nbd7'
    subprocess.check_call(['modprobe', 'nbd', 'max_part=64'])
    subprocess.check_call(['qemu-nbd', '--connect=' + device, disk_image])
    subprocess.check_call(['partprobe', device])
    output = subprocess.check_output(['fdisk', '-o', 'Device', '-l', device])
    root_device = output.decode().split('\n')[-2]

    return {
        'root_device': root_device,
        'image_type': 'vm',
        'mapped_device': device
    }


def map_raw_disk_image(disk_image):
    """Map the partitions inside a raw disk image as block devices."""
    logger.info('Adding partition mappings for raw disk image - %s',
                disk_image)
    output = subprocess.check_output(['kpartx', '-a', '-v', '-s', disk_image])
    output = output.decode()

    devices = []
    for line in output.split('\n'):
        if line:
            devices.append(line.split(' ')[2])

    root_device = '/dev/mapper/' + devices[-1]
    return {
        'root_device': root_device,
        'image_type': 'raw',
        'mapped_image': disk_image
    }


def mount_disk_image(root_device):
    """Mount the root device into a temporary directory and return the path."""
    mount_path = tempfile.mkdtemp()

    logger.info('Mounting %s on %s', root_device, mount_path)
    subprocess.check_call(['mount', root_device, mount_path])

    mount_info = {'mount_path': mount_path, 'root_path': mount_path}

    # XXX: Assumption that if root/@ exists, it is btrfs and that we
    # are going use that snapshot to work on it.
    if os.path.isdir(os.path.join(mount_path, '@')):
        mount_info['root_path'] = os.path.join(mount_path, '@')

    return mount_info


def change_password(root_path, user, password):
    """Change a user's password inside chroot directory."""
    logger.info('Changing password for %s inside %s', user, root_path)
    chpasswd_input = '{0}:{1}'.format(user, password)

    # XXX: Providing crypt method is not recommended.  However, without crypt
    # method, the passwd encryption happens using PAM and that does not seem to
    # be working in a chroot.
    subprocess.check_output(
        ['chpasswd', '--root', root_path, '--crypt-method', 'SHA512'],
        input=chpasswd_input.encode())


def unmount_disk_image(mount_info):
    """Unmount the root device."""
    mount_path = mount_info['mount_path']

    logger.info('Unmounting %s', mount_path)
    subprocess.check_call(['umount', mount_path])

    os.rmdir(mount_path)

    return mount_path


def unmap_disk_image(map_info):
    """Ummap the disk image partitions."""
    if map_info['image_type'] == 'vm':
        unmap_vm_disk_image(map_info)
    else:
        unmap_raw_disk_image(map_info)


def unmap_vm_disk_image(map_info):
    """Ummap the VM disk image partitions."""
    device = map_info['mapped_device']
    logger.info('Removing partition mappings from VM device - %s', device)
    subprocess.check_output(['qemu-nbd', '--disconnect', device])


def unmap_raw_disk_image(map_info):
    """Ummap the raw disk image partitions."""
    disk_image = map_info['mapped_image']
    logger.info('Removing partition mappings from raw disk - %s', disk_image)
    subprocess.check_output(['kpartx', '-d', disk_image])


if __name__ == '__main__':
    main()



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Tests for checking Freedom Maker's internal library of actions.
"""

import contextlib
import logging
import os
import random
import stat
import string
import subprocess
import tempfile
from unittest.mock import Mock, call, patch

import pytest

from .. import library, releases


@pytest.fixture(name='random_string')
def fixture_random_string():
    """Generate a random string."""

    def _random_string():
        return ''.join(
            [random.choice(string.ascii_lowercase) for _ in range(8)])

    return _random_string


@pytest.fixture(name='image')
def fixture_image(random_string):
    """Return an image file name."""
    return random_string()


@pytest.fixture(name='state')
def fixture_state(image):
    """Create a temporary mount point and return mid-build state."""
    with tempfile.TemporaryDirectory() as mount_point_directory:
        os.makedirs(mount_point_directory + '/usr/sbin/')
        os.makedirs(mount_point_directory + '/tmp/')
        os.makedirs(mount_point_directory + '/etc/apt')

        yield {
            'image_file': image,
            'mount_point': mount_point_directory,
            'success': True,
        }


@contextlib.contextmanager
def assert_file_change(path, content, expected_content):
    """Context manager to verify that file contents changed as expected."""
    if content is not None:
        with open(path, 'w', encoding='utf-8') as file_handle:
            file_handle.write(content)

    yield

    with open(path, 'r', encoding='utf-8') as file_handle:
        changed_content = file_handle.read()

    assert expected_content == changed_content


def test_run(caplog):
    """Test the run utility."""

    # Environment passing
    expected_environ = {
        'LC_ALL': 'C',
        'LANGUAGE': 'C',
        'LANG': 'C',
        'DEBIAN_FRONTEND': 'noninteractive',
        'DEBCONF_NONINTERACTIVE_SEEN': 'true'
    }
    output = library.run(['bash', '-c', 'set'])
    lines = output.splitlines()
    for key, value in expected_environ.items():
        assert f'{key}={value}'.encode() in lines

    # Argument passing
    output = library.run(['bash', '-c', 'echo $0 $@', '1', '2', '3', '4', '5'])
    assert output == b'1 2 3 4 5\n'

    # Stdout printing
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    library.run(['bash', '-c', 'echo "Hello"'])
    assert caplog.record_tuples == [
        ('freedommaker.library', logging.INFO, 'Executing command - '
         "['bash', '-c', 'echo \"Hello\"'] {}"),
        ('freedommaker.library', logging.DEBUG, '> Hello')
    ]

    # Stderr printing
    caplog.clear()
    library.run(['bash', '-c', 'echo "Hello" 1>&2'])
    assert caplog.record_tuples == [
        ('freedommaker.library', logging.INFO, 'Executing command - '
         "['bash', '-c', 'echo \"Hello\" 1>&2'] {}"),
        ('freedommaker.library', logging.DEBUG, '2> Hello')
    ]

    # Return value
    library.run(['bash', '-c', 'exit 0'], check=True)
    library.run(['bash', '-c', 'exit 0'], check=False)
    library.run(['bash', '-c', 'exit 1'], check=False)
    with pytest.raises(subprocess.CalledProcessError):
        library.run(['bash', '-c', 'exit 1'])


@patch('freedommaker.library.run')
def test_run_in_chroot(run, state):
    """Test executing inside a chroot environment."""
    args = ['1', '2', '3']
    kwargs = {'a': 'x', 'b': 'y'}

    library.run_in_chroot(state, args, **kwargs)
    expected_args = ['chroot', state['mount_point']] + args
    assert run.call_args == call(expected_args, **kwargs)


def test_path_in_mount(state):
    """Test returning a sub-directory in mount point."""
    output = library.path_in_mount(state, 'boot')
    assert output == state['mount_point'] + '/boot'

    output = library.path_in_mount(state, '/boot')
    assert output == '/boot'


def test_schedule_clean(state):
    """Test scheduling cleanup jobs."""
    cleanup = Mock()
    library.schedule_cleanup(state, cleanup, 2, b=3)
    assert state['cleanup'][-1] == [cleanup, (2, ), {'b': 3}]


def test_cleanup(state):
    """Test the cleanup works."""
    method = Mock()
    library.schedule_cleanup(state, method, 1, a=1)
    library.schedule_cleanup(state, method, 2, a=2)
    library.schedule_cleanup(state, method, 3, a=3)
    library.cleanup(state)
    assert method.call_args_list == [((3, ), {
        'a': 3
    }), ((2, ), {
        'a': 2
    }), ((1, ), {
        'a': 1
    })]


@patch('freedommaker.library.run')
def test_create_ram_directory_image(run, image, state):
    """Test that RAM directory is properly created."""
    library.create_ram_directory_image(state, image, '4G')
    run.assert_called_once_with([
        'mount', '-o', 'size=4G', '-t', 'tmpfs', 'tmpfs',
        state['ram_directory'].name
    ])
    assert state['image_file'] == (state['ram_directory'].name + '/' + image)
    assert state['cleanup'] == [[
        library.remove_ram_directory, (state['ram_directory'], ), {}
    ], [library.copy_image, (state, state['image_file'], image), {}]]
    state['ram_directory'].cleanup()


@patch('freedommaker.library.run')
def test_remove_ram_directory(run, random_string):
    """Test removing a RAM directory."""
    directory = Mock(name=random_string())
    library.remove_ram_directory(directory)
    run.assert_called_once_with(['umount', directory.name])
    directory.cleanup.assert_called()


@patch('freedommaker.library.run')
def test_copy_image(run, random_string, image, state):
    """Test copying temp image to final image."""
    temp_image = random_string()
    library.copy_image(state, temp_image, image)
    run.assert_called_once_with(['cp', '--sparse=always', temp_image, image])

    run.reset_mock()
    state['success'] = False
    library.copy_image(state, temp_image, image)
    run.assert_called_once_with(
        ['cp', '--sparse=always', temp_image, image + '.failed'])


def test_create_temp_image(image, state):
    """Test creating a temporary image file on disk."""
    library.create_temp_image(state, image)
    assert state['image_file'] == image + '.temp'
    assert state['cleanup'] == [[
        library.move_image, (state, image + '.temp', image), {}
    ]]


@patch('freedommaker.library.run')
def test_move_image(run, random_string, image, state):
    """Test moving temp image to final image."""
    source_image = random_string()
    library.move_image(state, source_image, image)
    run.assert_called_once_with(['mv', source_image, image])

    run.reset_mock()
    state['success'] = False
    library.move_image(state, source_image, image)
    run.assert_called_once_with(['mv', source_image, image + '.failed'])


@patch('freedommaker.library.run')
def test_create_image(run, image, state):
    """Test creating an image."""
    library.create_image(state, '4G')
    run.assert_called_once_with(
        ['qemu-img', 'create', '-f', 'raw', image, '4G'])


@patch('freedommaker.library.run')
def test_create_partition_table(run, image, state):
    """Test creating a partition table."""
    library.create_partition_table(state, 'msdos')
    run.assert_called_once_with(['parted', '-s', image, 'mklabel', 'msdos'])


@patch('freedommaker.library.run')
def test_create_partition(run, image, state):
    """Test creating a partition table."""
    library.create_partition(state, 'root', '10mib', '50%', 'f2fs')
    run.assert_called_once_with(
        ['parted', '-s', image, 'mkpart', 'primary', 'f2fs', '10mib', '50%'])

    assert state['partitions'] == ['root']

    library.create_partition(state, 'root', '10mib', '50%', 'vfat')
    run.assert_called_with(
        ['parted', '-s', image, 'mkpart', 'primary', 'fat32', '10mib', '50%'])


@patch('freedommaker.library.run')
def test_set_boot_flag(run, image, state):
    """Test that boot flag is properly set."""
    library.set_boot_flag(state, 3)
    run.assert_called_with(['parted', '-s', image, 'set', '3', 'boot', 'on'])


@patch('freedommaker.library.run')
def test_loopback_setup(run, image, state):
    """Test that loopback device is properly setup."""
    state['partitions'] = ['firmware', 'boot', 'root']

    run.return_value = b'''remove x x
add x loop99p1
add x loop99p2
add x loop99p3
modify x x
'''
    library.loopback_setup(state)
    run.assert_called_with(['kpartx', '-asv', image])
    assert state['devices'] == {
        'firmware': '/dev/mapper/loop99p1',
        'boot': '/dev/mapper/loop99p2',
        'root': '/dev/mapper/loop99p3'
    }
    assert state['loop_device'] == '/dev/loop99'
    assert state['cleanup'] == [
        [library.force_release_loop_device, ('/dev/loop99', ), {}],
        [library.force_release_partition_loop, ('/dev/mapper/loop99p1', ), {}],
        [library.force_release_partition_loop, ('/dev/mapper/loop99p2', ), {}],
        [library.force_release_partition_loop, ('/dev/mapper/loop99p3', ), {}],
        [library.loopback_teardown, (image, ), {}]
    ]


@patch('freedommaker.library.run')
def test_force_release_partition_loop(run):
    """Test loop device is forcefully released."""
    library.force_release_partition_loop('/dev/test/loop99')
    run.assert_called_with(['dmsetup', 'remove', '/dev/test/loop99'],
                           check=False)


@patch('freedommaker.library.run')
def test_force_release_loop_device(run):
    """Test loop device is forcefully released."""
    library.force_release_partition_loop('/dev/test/loop99')
    run.assert_called_with(['dmsetup', 'remove', '/dev/test/loop99'],
                           check=False)


@patch('freedommaker.library.run')
def test_loopback_teardown(run):
    """Test tearing down of loopback."""
    library.loopback_teardown('/dev/test/loop99')
    run.assert_called_with(['kpartx', '-dsv', '/dev/test/loop99'])


@patch('freedommaker.library.run')
def test_create_filesystem(run):
    """Test creating filesystem."""
    library.create_filesystem('/dev/test/loop99p1', 'btrfs')
    assert run.call_args_list == [
        call(['mkfs', '-t', 'btrfs', '/dev/test/loop99p1']),
        call(['udevadm', 'trigger', '/dev/test/loop99p1']),
        call(['udevadm', 'settle'])
    ]


@patch('freedommaker.library.run')
def test_mount_filesystem(run, state):
    """Test mounting a filesystem and setting proper state."""
    state['devices'] = {
        'root': '/dev/test/loop99p1',
        'firmware': '/dev/test/loop99p2'
    }
    library.mount_filesystem(state, 'root', None)
    run.assert_called_with(
        ['mount', '/dev/test/loop99p1', state['mount_point']])

    library.mount_filesystem(state, 'firmware', 'boot/firmware')
    run.assert_called_with([
        'mount', '/dev/test/loop99p2', state['mount_point'] + '/boot/firmware'
    ])

    library.mount_filesystem(state, '/dev/pts', 'dev/pts', is_bind_mount=True)
    run.assert_called_with(
        ['mount', '/dev/pts', state['mount_point'] + '/dev/pts', '-o', 'bind'])

    sub_mount_points = {
        'root': None,
        'firmware': 'boot/firmware',
        '/dev/pts': 'dev/pts'
    }
    assert state['sub_mount_points'] == sub_mount_points


@patch('freedommaker.library.run')
def test_unmount_filesystem(run, state):
    """Test unmounting a filesystem."""
    library.unmount_filesystem('/dev/', state['mount_point'], True)
    assert run.call_args_list == [
        call(['umount', state['mount_point']], check=True)
    ]


@patch('freedommaker.library.run')
def test_process_cleanup(run, state):
    """Test cleaning up processes."""
    library.process_cleanup(state)
    assert run.call_args_list == [
        call(['fuser', '-mvk', state['mount_point']], check=False),
        call(['fuser', '-mvk', state['mount_point']], check=False)
    ]


@patch('freedommaker.library.run')
def test_setup_extra_storage(run, image, state, random_string):
    """Test that setting up extra storage works."""
    extra_storage_file = image + '.extra'
    size = '100M'
    mount_point = state['mount_point']
    loop_device = random_string()
    run.return_value = f'{loop_device}\n'.encode()

    library.setup_extra_storage(state, 'ext4', '100M')
    run.assert_not_called()

    library.setup_extra_storage(state, 'btrfs', '100M')
    run.assert_has_calls([
        call(['qemu-img', 'create', '-f', 'raw', extra_storage_file, size]),
        call(['losetup', '--show', '--find', extra_storage_file]),
        call(['btrfs', 'device', 'add', loop_device, mount_point])
    ])

    assert state['cleanup'] == [[
        library.cleanup_extra_storage,
        (state, loop_device, extra_storage_file), {}
    ]]


@patch('freedommaker.library.run')
def test_cleanup_extra_storage(run, image, state, random_string):
    """Test that extra storage will be cleaned properly."""
    extra_storage_file = image + '.extra'
    loop_device = random_string()
    mount_point = state['mount_point']

    library.cleanup_extra_storage(state, loop_device, extra_storage_file)

    run.assert_has_calls([
        call(['btrfs', 'balance', 'start', '-musage=0', mount_point],
             check=False),
        call(['btrfs', 'balance', 'start', '-dusage=0', mount_point],
             check=False),
        call(['btrfs', 'balance', 'start', '-musage=20', mount_point],
             check=False),
        call(['btrfs', 'balance', 'start', '-dusage=20', mount_point],
             check=False),
        call(['btrfs', 'balance', 'start', '-musage=40', mount_point],
             check=False),
        call(['btrfs', 'balance', 'start', '-dusage=40', mount_point],
             check=False),
        call(['btrfs', 'balance', 'start', '-musage=60', mount_point],
             check=False),
        call(['btrfs', 'balance', 'start', '-dusage=60', mount_point],
             check=False),
        call(['btrfs', 'balance', 'start', '-musage=80', mount_point],
             check=False),
        call(['btrfs', 'balance', 'start', '-dusage=80', mount_point],
             check=False),
        call(['btrfs', 'balance', 'start', '-mconvert=dup', mount_point],
             check=False),
        call(['btrfs', 'device', 'remove', loop_device, mount_point]),
        call(['losetup', '--detach', loop_device]),
        call(['rm', '-f', extra_storage_file]),
        call(['btrfs', 'balance', 'start', '-musage=0', mount_point],
             check=False),
        call(['btrfs', 'balance', 'start', '-dusage=0', mount_point],
             check=False),
        call(['btrfs', 'balance', 'start', '-musage=20', mount_point],
             check=False),
        call(['btrfs', 'balance', 'start', '-dusage=20', mount_point],
             check=False),
        call(['btrfs', 'balance', 'start', '-musage=40', mount_point],
             check=False),
        call(['btrfs', 'balance', 'start', '-dusage=40', mount_point],
             check=False),
        call(['btrfs', 'balance', 'start', '-musage=60', mount_point],
             check=False),
        call(['btrfs', 'balance', 'start', '-dusage=60', mount_point],
             check=False),
        call(['btrfs', 'balance', 'start', '-musage=80', mount_point],
             check=False),
        call(['btrfs', 'balance', 'start', '-dusage=80', mount_point],
             check=False),
    ])


@patch('freedommaker.library.run')
def test_debootstrap(run, state):
    """Test debootstrapping."""
    library.debootstrap(state, 'amd64', 'stretch', 'minbase',
                        ['main', 'contrib'], ['p1', 'p2'],
                        'http://deb.debian.org/debian')
    run.assert_called_with([
        'debootstrap', '--arch=amd64', '--variant=minbase',
        '--components=main,contrib', '--include=p1,p2', 'stretch',
        state['mount_point'], 'http://deb.debian.org/debian'
    ])

    assert state['cleanup'] == [[
        library.unmount_filesystem,
        (None, state['mount_point'] + '/etc/machine-id'), {
            'check': False
        }
    ]]


def test_no_daemon_policy(state):
    """Test that no daemon run policy is properly set."""
    file_path = state['mount_point'] + '/usr/sbin/policy-rc.d'
    with library.no_run_daemon_policy(state):
        with open(file_path, 'r', encoding='utf-8') as file_handle:
            contents = file_handle.read()

        assert contents == '#!/bin/sh\nexit 101\n'
        assert oct(os.stat(file_path)[stat.ST_MODE])[-3:] == '755'

        assert os.path.isfile(file_path)


@patch('freedommaker.library.run_in_chroot')
def test_install_package(run, state):
    """Test installing a package."""
    library.install_package(state, 'nmap')
    run.assert_called_with(state, ['apt-get', 'install', '-y', 'nmap'])


@patch('freedommaker.library.run_in_chroot')
def test_install_from_backports(run, state):
    """Test installing a package from backports."""
    library.install_from_backports(
        state, ('freedombox', 'freedombox-doc-en', 'freedombox-doc-es'))

    assert run.call_args_list == [
        call(state, [
            'apt-get', 'install', '-y',
            f'freedombox/{releases.STABLE_CODENAME}-backports',
            f'freedombox-doc-en/{releases.STABLE_CODENAME}-backports',
            f'freedombox-doc-es/{releases.STABLE_CODENAME}-backports'
        ]),
        call(state, ['apt-get', 'autoremove', '-y'])
    ]


@patch('freedommaker.library.install_package')
@patch('freedommaker.library.run_in_chroot')
def test_install_custom_package(run, _install_package, state):
    """Test installing a custom package."""
    with tempfile.NamedTemporaryFile() as file_path:
        library.install_custom_package(state, file_path.name)

        run.assert_called_with(state, [
            'apt', 'install', '-y', '--allow-downgrades',
            '/tmp/' + os.path.basename(file_path.name)
        ])


def test_set_hostname(state):
    """Test that hostname is properly set."""
    hosts_path = state['mount_point'] + '/etc/hosts'
    hostname_path = state['mount_point'] + '/etc/hostname'

    content = '''127.0.0.1 localhost
::1     localhost ip6-localhost ip6-loopback
ff02::1 ip6-allnodes
ff02::2 ip6-allrouters
'''
    expected_content = content + '''127.0.1.1 fbx\n'''
    with assert_file_change(hosts_path, content, expected_content):
        with assert_file_change(hostname_path, 'test', 'fbx\n'):
            library.set_hostname(state, 'fbx')


def test_get_fstab_options():
    """Test getting fstab options for a file system."""
    options = library.get_fstab_options('ext4')
    assert options == 'errors=remount-ro'

    options = library.get_fstab_options('btrfs')
    assert options == 'compress=zstd:1'


@patch('freedommaker.library.run', return_value=b'test-uuid')
def test_get_uuid_of_device(run):
    """Test retrieving UUID of device."""
    response = library.get_uuid_of_device('/dev/test/loop99p1')
    run.assert_called_with(
        ['blkid', '--output=value', '--match-tag=UUID', '/dev/test/loop99p1'])
    assert response == 'test-uuid'


@patch('freedommaker.library.get_uuid_of_device')
def test_add_fstab_entry(get_uuid, state):
    """Test adding entries to /etc/fstab."""
    fstab_path = state['mount_point'] + '/etc/fstab'
    state['devices'] = {
        'root': '/dev/test/loop99p1',
        'boot': '/dev/test/loop99p2'
    }
    state['sub_mount_points'] = {'root': None, 'boot': 'boot'}

    expected_content = 'UUID=root-uuid / btrfs compress=zstd:1 0 1\n'
    get_uuid.return_value = 'root-uuid'
    with assert_file_change(fstab_path, 'initial-trash', expected_content):
        library.add_fstab_entry(state, 'root', 'btrfs', 1, append=False)

    expected_content += 'UUID=boot-uuid /boot ext4 errors=remount-ro 0 2\n'
    get_uuid.return_value = 'boot-uuid'
    with assert_file_change(fstab_path, None, expected_content):
        library.add_fstab_entry(state, 'boot', 'ext4', 2, append=True)


@patch('freedommaker.library.run_in_chroot')
def test_install_grub(run, state):
    """Test installing grub boot loader."""
    state['loop_device'] = '/dev/test/loop99'
    library.install_grub(state)
    assert run.call_args_list == [
        call(state, ['update-grub']),
        call(state, ['grub-install', '/dev/test/loop99'])
    ]

    run.reset_mock()
    library.install_grub(state, target='arm-efi')
    assert run.call_args_list == [
        call(state, ['update-grub']),
        call(state, ['grub-install', '/dev/test/loop99', '--target=arm-efi'])
    ]

    run.reset_mock()
    library.install_grub(state, target='arm-efi', is_efi=True)
    assert run.call_args_list == [
        call(state, ['update-grub']),
        call(state, [
            'grub-install', '/dev/test/loop99', '--target=arm-efi',
            '--no-nvram'
        ])
    ]


@patch('freedommaker.library.run_in_chroot')
def test_setup_apt(run, state):
    """Test setting up apt."""
    sources_path = state['mount_point'] + '/etc/apt/sources.list'

    stable_content = '''
deb http://deb.debian.org/debian stable main
deb-src http://deb.debian.org/debian stable main

deb http://deb.debian.org/debian stable-updates main
deb-src http://deb.debian.org/debian stable-updates main

deb http://security.debian.org/debian-security/ stable-security main
deb-src http://security.debian.org/debian-security/ stable-security main
'''
    with assert_file_change(sources_path, None, stable_content):
        library.setup_apt(state, 'http://deb.debian.org/debian', 'stable',
                          ['main'])

    assert run.call_args_list == [
        call(state, ['apt-get', 'update']),
        call(state, ['apt-get', 'clean'])
    ]

    unstable_content = '''
deb http://ftp.us.debian.org/debian unstable main contrib non-free-firmware
deb-src http://ftp.us.debian.org/debian unstable main contrib non-free-firmware
'''
    with assert_file_change(sources_path, None, unstable_content):
        library.setup_apt(state, 'http://ftp.us.debian.org/debian', 'unstable',
                          ['main', 'contrib', 'non-free-firmware'])


@patch('freedommaker.library.run_in_chroot')
def test_setup_flash_kernel(run, state):
    """Test setting up flash kernel."""
    machine_path = state['mount_point'] + '/etc/flash-kernel/machine'
    expected_content = 'test-machine'
    with assert_file_change(machine_path, None, expected_content):
        library.setup_flash_kernel(state, 'test-machine', None, 'ext2')

    assert run.call_args_list == [
        call(state, ['apt-get', 'install', '-y', 'flash-kernel']),
        call(state, ['flash-kernel'])
    ]

    run.reset_mock()
    with assert_file_change(machine_path, None, expected_content):
        library.setup_flash_kernel(state, 'test-machine', 'debug', 'vfat')

    selection = b'flash-kernel flash-kernel/linux_cmdline string debug'
    assert run.call_args_list == [
        call(state, ['debconf-set-selections'], input=selection),
        call(state, ['apt-get', 'install', '-y', 'flash-kernel']),
    ]


@patch('freedommaker.library.run_in_chroot')
def test_update_initramfs(run, state):
    """Test updating initramfs."""
    library.update_initramfs(state)

    assert run.call_args_list == [call(state, ['update-initramfs', '-u'])]


@patch('freedommaker.library.run')
def test_install_boot_loader_path(run, image, state):
    """Test installing boot loader components using dd."""
    path = 'u-boot/path'
    full_path = state['mount_point'] + '/' + path
    library.install_boot_loader_part(state, path, '533', '515')
    run.assert_called_with([
        'dd', 'if=' + full_path, 'of=' + image, 'seek=533', 'bs=515',
        'conv=notrunc'
    ])

    library.install_boot_loader_part(state, path, '533', '515', '90')
    run.assert_called_with([
        'dd', 'if=' + full_path, 'of=' + image, 'seek=533', 'bs=515',
        'conv=notrunc', 'count=90'
    ])


@patch('freedommaker.library.run')
def test_fill_free_space_with_zeros(run, state):
    """Test filling free space with zeros."""
    zeros_path = state['mount_point'] + '/ZEROS'
    library.fill_free_space_with_zeros(state)
    assert run.call_args_list == [
        call(['dd', 'if=/dev/zero', 'of=' + zeros_path, 'bs=1M'], check=False),
        call(['rm', '-f', zeros_path])
    ]


@patch('freedommaker.library.run')
def test_compress(run, random_string):
    """Test compressing an image."""
    archive_file = random_string()
    image_file = random_string()

    library.compress(archive_file, image_file)
    run.assert_called_with([
        'xz', '--no-warn', '--threads=0', '--memlimit=50%', '-9', '--force',
        image_file
    ])


@patch('os.remove')
@patch('freedommaker.library.run')
def test_sign(run, remove, random_string):
    """Test running signing with GPG."""
    archive = random_string()
    signature = archive + '.sig'
    remove.side_effect = FileNotFoundError
    library.sign(archive)
    run.assert_called_with(
        ['gpg', '--output', signature, '--detach-sig', archive])



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Worker class to build AMD64 images.
"""

from .amd_intel import AMDIntelImageBuilder


class AMD64ImageBuilder(AMDIntelImageBuilder):
    """Image builder for all amd64 targets."""
    architecture = 'amd64'
    kernel_flavor = 'amd64'
    include_non_free_firmware = True

    def __init__(self, *args, **kwargs):
        """Initialize builder object."""
        super().__init__(*args, **kwargs)
        self.packages += ['amd64-microcode', 'intel-microcode']



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Worker class to build Raspberry Pi 3 image.
"""

from .raspberry_pi_with_uboot import RaspberryPiWithUBoot


class RaspberryPi3ImageBuilder(RaspberryPiWithUBoot):
    """Image builder for Raspberry Pi 3 target."""
    architecture = 'armhf'
    machine = 'raspberry3'
    kernel_flavor = 'armmp'
    flash_kernel_name = 'Raspberry Pi 3 Model B'
    u_boot_rpi_variant = 'rpi_3_32b'



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Information about various Debian releases.
"""

STABLE_CODENAME = 'bookworm'

STABLE_DISTRIBUTIONS = ['stable', 'bullseye', 'bookworm']


def is_stable(distribution: str):
    """Return whether the selected distribution is stable."""
    return distribution in STABLE_DISTRIBUTIONS



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Worker class to build LinkSprite pcDuino3 image.
"""

from .a20 import A20ImageBuilder


class PCDuino3ImageBuilder(A20ImageBuilder):
    """Image builder for PCDuino3 target."""
    machine = 'pcduino3'
    flash_kernel_name = 'LinkSprite pcDuino3'
    u_boot_path = \
        'usr/lib/u-boot/Linksprite_pcDuino3/u-boot-sunxi-with-spl.bin'



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Worker class to build Raspberry Pi 64-bit image.
"""

from .. import library
from ..builder import ImageBuilder


class RaspberryPi64ImageBuilder(ImageBuilder):
    """Image builder for Raspberry Pi 64-bit target."""
    architecture = 'arm64'
    boot_loader = None
    firmware_filesystem_type = 'vfat'
    firmware_size = '256MiB'
    include_non_free_firmware = True
    kernel_flavor = 'arm64'
    machine = 'raspberry64'
    update_initramfs = False

    def __init__(self, arguments):
        """Add to list of packages."""
        super().__init__(arguments)
        self.packages += ['firmware-brcm80211']

    @classmethod
    def get_target_name(cls):
        """Return the name of the target for an image builder."""
        return getattr(cls, 'machine', None)

    def install_boot_loader(self, state):
        """Install the firmware onto the image."""
        uuid = library.get_uuid_of_device(state['devices']['root'])
        script = '''
set -e
set -x
set -o pipefail

# This will trigger installing firmware and updating initramfs.
apt-get install raspi-firmware

# Fixup cmdline.txt and set defaults
echo "console=tty0 console=ttyS1,115200 root=UUID={uuid} rw fsck.repair=yes net.ifnames=0 rootwait" >/boot/firmware/cmdline.txt
sed -i 's/^#ROOTPART=.*/ROOTPART=UUID={uuid}/' /etc/default/raspi-firmware
'''.format(uuid=uuid)
        library.run_in_chroot(state, ['bash', '-c', script])



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Worker class to build Olimex A20-OLinuXino Micro image.
"""

from .a20 import A20ImageBuilder


class A20OLinuXinoMicroImageBuilder(A20ImageBuilder):
    """Image builder for A20 OLinuXino Micro targets."""
    machine = 'a20-olinuxino-micro'
    flash_kernel_name = 'Olimex A20-Olinuxino Micro'
    u_boot_path = \
        'usr/lib/u-boot/A20-OLinuXino_MICRO/u-boot-sunxi-with-spl.bin'



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Worker class to build LeMaker Banana Pro image.
"""

from .a20 import A20ImageBuilder


class BananaProImageBuilder(A20ImageBuilder):
    """Image builder for Banana Pro target."""
    machine = 'banana-pro'
    flash_kernel_name = 'LeMaker Banana Pro'
    u_boot_path = 'usr/lib/u-boot/Bananapro/u-boot-sunxi-with-spl.bin'



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Base worker class to build all ARM64 images.
"""

import os

from .. import library
from .arm import ARMImageBuilder


class A64ImageBuilder(ARMImageBuilder):
    """Image builder for all Allwinner A64 board targets."""
    architecture: str = 'arm64'
    kernel_flavor: str = 'arm64'
    u_boot_target: str | None = None

    def __init__(self, *args, **kwargs):
        """Initialize builder object."""
        super().__init__(*args, **kwargs)
        self.packages += ['arm-trusted-firmware', 'device-tree-compiler']

    @classmethod
    def install_boot_loader(cls, state):
        """Install the boot loader onto the image."""
        if not cls.u_boot_target:
            raise ValueError('U-boot target must be provided.')

        u_boot_target_path = '/usr/lib/u-boot/{}'.format(cls.u_boot_target)
        env = os.environ.copy()
        env['TARGET'] = u_boot_target_path

        library.run_in_chroot(state,
                              ['u-boot-install-sunxi64', state['loop_device']],
                              env=env)



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Worker class to build Vagrant images.
"""

import logging
import os
import pathlib
import subprocess

from .. import library
from .amd64 import AMD64ImageBuilder

logger = logging.getLogger(__name__)

vagrant_ssh_keys = 'https://raw.githubusercontent.com/mitchellh/vagrant/' \
    'master/keys/vagrant.pub'


class VagrantImageBuilder(AMD64ImageBuilder):
    """Image builder for Vagrant package."""
    include_contrib = True
    vagrant_extension = '.box'

    @classmethod
    def get_target_name(cls):
        """Return the name of the target for an image builder."""
        return 'vagrant'

    def build(self):
        """Run the image building process."""
        # Create a .img hard disk image
        self.make_image()

        # Convert hard disk image from .img file to .vmdk file
        build_path = pathlib.Path(self.arguments.build_dir)
        vm_file = build_path / 'box-disk001.vmdk'
        self._create_vm_file(self.image_file, vm_file)
        os.remove(self.image_file)

        # Create metadata.json
        metadata_file = build_path / 'metadata.json'
        metadata_file.write_text('{"provider":"virtualbox"}')

        # Create Vagrantfile from a template in ./vagrant/ directory
        vagrant_file = pathlib.Path(self.arguments.build_dir) / 'Vagrantfile'
        vagrant_text = self._read_template('Vagrantfile')
        vagrant_file.write_text(vagrant_text)

        # Create box.ovf from a template in ./vagrant/ directory
        ovf_file = pathlib.Path(self.arguments.build_dir) / 'box.ovf'
        ovf_text = self._read_template('box.ovf')
        ovf_file.write_text(ovf_text)

        # Create a vagrant .box file for upload
        box_file = pathlib.Path(self.image_file).with_suffix(
            self.vagrant_extension)
        self._create_vagrant_box(box_file, vm_file, metadata_file,
                                 vagrant_file, ovf_file)

        # Create a hash file
        self._store_hash(box_file)

    def extra_setup(self, state):
        """Run additional setup needed by vagrant. Called from make_image()."""
        if self.arguments.distribution in ['unstable', 'sid']:
            # XXX: Only unstable will have VirtualBox's shared folder support
            # and time synchronization service.
            self._install_guest_additions(state, self.release_components)

        if self.arguments.distribution not in [
                'stable', 'bullseye', 'bookworm'
        ]:
            # XXX: Stable will not have packages required to build freedombox
            # package. This requires latest debhelper version, which can be
            # installed from backports.
            self._install_dev_packages(state)

        self._setup_vagrant_user(state)

    def _install_guest_additions(self, state, release_components):
        """Install VirtualBox Guest Additions into the VM image."""
        if 'contrib' not in release_components:
            logger.warning(
                'Skipping installation of VirtualBox Guest Additions.')
            logger.warning('virtualbox-guest-utils is only available with the '
                           '"contrib" release component.')
            logger.warning(
                'The following release components were used in this '
                'build: ' + ' '.join(release_components))
            return

        library.install_package(state, 'virtualbox-guest-utils')

    def _install_dev_packages(self, state):
        """Install build deps and other useful packages for development."""
        logger.info('Installing freedombox build dependencies.')
        library.run_in_chroot(state,
                              ['apt-get', 'build-dep', '-y', 'freedombox'])

        logger.info('Installing extra development utilities.')
        packages = [
            'byobu', 'ncurses-term', 'parted', 'python3-dev', 'python3-pip',
            'python3-pytest'
        ]
        library.run_in_chroot(state, ['apt-get', 'install', '-y'] + packages)

    def _setup_vagrant_user(self, state):
        """Create vagrant user, setup SSH keys, groups, and sudo."""
        logger.info('Setting up vagrant user.')

        def _in_chroot(command):
            library.run_in_chroot(state, command)

        _in_chroot(
            ['adduser', '--disabled-password', '--gecos', '', 'vagrant'])

        _in_chroot(['mkdir', '/home/vagrant/.ssh'])
        _in_chroot([
            'wget', '-O', '/home/vagrant/.ssh/authorized_keys',
            vagrant_ssh_keys
        ])
        _in_chroot(['chown', '-R', 'vagrant:vagrant', '/home/vagrant/.ssh'])
        _in_chroot(['chmod', '0700', '/home/vagrant/.ssh'])
        _in_chroot(['chmod', '0600', '/home/vagrant/.ssh/authorized_keys'])

        _in_chroot(['usermod', '-a', '-G', 'sudo', 'vagrant'])
        file_path = pathlib.Path(
            library.path_in_mount(state, 'etc/sudoers.d/vagrant'))
        file_path.write_text('vagrant ALL=(ALL) NOPASSWD: ALL')

    def _create_vm_file(self, image_file, vm_file):
        """Convert the disk image into a virtual machine image."""
        library.run(['qemu-img', 'convert', '-O', 'vmdk', image_file, vm_file])

    def _read_template(self, file_name):
        """Read a template file for a constituent of vagrant .box file."""
        file_path = pathlib.Path(__file__).parent / 'vagrant' / file_name
        return file_path.read_text()

    def _create_vagrant_box(self, box_file, vm_file, metadata_file,
                            vagrant_file, ovf_file):
        """Create a vagrant .box file from constituents."""
        library.run([
            'tar', '-cvzf', box_file.name, vm_file.name, metadata_file.name,
            vagrant_file.name, ovf_file.name
        ],
                    cwd=vm_file.parent)

    def _store_hash(self, box_file):
        """Store the SHA-256 hash of the vagrant box file."""
        output = subprocess.check_output(['sha256sum', box_file])
        result = output.decode()
        logger.info('sha256sum: %s', result)
        hash_filename = box_file.with_suffix('.sha256')
        with open(hash_filename, 'w') as hash_file:
            hash_file.write(result)



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Worker class to build Xunlong Orange Pi Zero image.
"""

from .a20 import A20ImageBuilder


class OrangePiZeroImageBuilder(A20ImageBuilder):
    """Image builder for Orange Pi Zero target."""
    machine = 'orange-pi-zero'
    flash_kernel_name = 'Xunlong Orange Pi Zero'
    u_boot_path = 'usr/lib/u-boot/orangepi_zero/u-boot-sunxi-with-spl.bin'



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Miscellaneous utilities that don't fit anywhere else.
"""

import re


def add_disk_offsets(offset1, offset2):
    """Add two disk offsets as understood by parted utility.

    Currently only adding offsets in MiB is supported. For parted 'mib' is
    different from 'MiB'. When specifying the ending sector of a partition,
    'mib' means that specified sector included like [start, end] and 'MiB'
    means that specified sector is not included like [start, end).

    """
    if not offset1.endswith('MiB') or not offset2.endswith('MiB'):
        raise NotImplementedError('Adding anything but offsets in MiB')

    result = int(offset1.rstrip('MiB')) + int(offset2.rstrip('MiB'))
    return '{}MiB'.format(result)


def parse_disk_size(input_size):
    """Return integer size for size strings like 1000M, 2G etc."""
    try:
        return int(input_size)
    except ValueError:
        pass

    if not re.fullmatch(r'\d+[bkKMGT]', input_size):
        raise ValueError('Invalid size: ' + input_size)

    size = input_size[:-1]
    unit = input_size[-1]
    size_map = {
        'b': 1,
        'k': 1024,
        'K': 1024,
        'M': 1024 * 1024,
        'G': 1024 * 1024 * 1024,
        'T': 1024 * 1024 * 1024 * 1024,
    }
    return int(size) * size_map[unit]


def format_disk_size(size):
    """Return string like 100M, 2G given size in bytes."""
    unit = 0
    while size >= 1024 and unit < 5 and int(size / 1024) * 1024 == size:
        unit += 1
        size = int(size / 1024)

    return str(size) + ['', 'K', 'M', 'G', 'T'][unit]


def add_disk_sizes(size1, size2):
    """Add two string sizes represented as 1000M, 2G, etc."""
    return format_disk_size(parse_disk_size(size1) + parse_disk_size(size2))



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Base worker class to build VirtualBox images.
"""

import os

from .. import library
from .vm import VMImageBuilder


class VirtualBoxImageBuilder(VMImageBuilder):
    """Base image builder for all VirtualBox targets."""
    vm_image_extension = '.vdi'

    @classmethod
    def get_target_name(cls):
        """Return the name of the target for an image builder."""
        if getattr(cls, 'architecture', None):
            return 'virtualbox-' + cls.architecture

        return None

    def create_vm_file(self, image_file, vm_file):
        """Create a VM file from image file."""
        try:
            os.remove(vm_file)
        except FileNotFoundError:
            pass

        library.run(['VBoxManage', 'convertdd', image_file, vm_file])



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Tests for checking built image with VirtualBox.
"""

import os
import pathlib
import random
import string
import subprocess
import tempfile
import time

import pytest


@pytest.fixture(name='random_string')
def fixture_random_string():
    """Generate a random string."""

    def _random_string():
        return ''.join(
            [random.choice(string.ascii_lowercase) for _ in range(8)])

    return _random_string


@pytest.fixture(name='build_info')
def fixture_build_info(random_string):
    """Create basic build information."""
    with tempfile.TemporaryDirectory() as output_dir:
        yield {
            'output_dir': output_dir,
            'build_stamp': random_string(),
            'current_target': 'virtualbox-amd64'
        }


@pytest.fixture(name='invoke')
def fixture_invoke(build_info):
    """Return a helper that invokes freedom-maker."""

    def _invoke(targets=None, **kwargs):
        """Invoke Freedom Maker."""
        parameters = ['--build-dir', build_info['output_dir']]

        if 'build_stamp' not in kwargs:
            parameters += ['--build-stamp', build_info['build_stamp']]

        command = ['python3', '-m', 'freedommaker'] + parameters + targets
        subprocess.check_call(command)

    return _invoke


def get_built_file(build_info):
    """Return the path of the expected built file."""
    file_name = 'freedombox-unstable_{build_stamp}_all-amd64.vdi.xz' \
        .format(build_stamp=build_info['build_stamp'])

    return os.path.join(build_info['output_dir'], file_name)


def _run(*args, ignore_errors=False):
    """Execute a command."""
    subprocess.run(*args, check=not ignore_errors)


@pytest.mark.skipif(os.environ.get('FM_RUN_VM_TESTS') != 'true',
                    reason='Not requested')
def test_basic_build(build_info, invoke):
    """Test booting and opening SSH shell.

    Also:
      - Output the plinth diagnostic log.
    """
    invoke(['virtualbox-amd64'])

    vm_name = 'freedom-maker-test'
    test_ssh_port = 2222
    first_run_wait_time = 120

    compressed_built_file = get_built_file(build_info)
    built_file = compressed_built_file.rsplit('.', maxsplit=1)[0]

    if not os.path.isfile(built_file):
        subprocess.check_call(
            ['unxz', '--keep', '--force', compressed_built_file])

    passwd_tool = pathlib.Path(__file__).parent / '../passwd_in_image.py'
    _run([
        'sudo', 'python3', passwd_tool, built_file, 'fbx', '--password', 'frdm'
    ])
    try:
        _run([
            'VBoxManage', 'createvm', '--name', vm_name, '--ostype', 'Debian',
            '--register'
        ])
        _run([
            'VBoxManage', 'storagectl', vm_name, '--name', 'SATA Controller',
            '--add', 'sata', '--controller', 'IntelAHCI'
        ])
        _run([
            'VBoxManage', 'storageattach', vm_name, '--storagectl',
            'SATA Controller', '--port', '0', '--device', '0', '--type', 'hdd',
            '--medium', built_file
        ])
        _run([
            'VBoxManage', 'modifyvm', vm_name, '--pae', 'on', '--memory',
            '1024', '--vram', '128', '--nic1', 'nat', '--natpf1',
            ',tcp,,{port},,22'.format(port=test_ssh_port)
        ])
        _run(['VBoxManage', 'startvm', vm_name, '--type', 'headless'])
        time.sleep(first_run_wait_time)

        echo = subprocess.Popen(['echo', 'frdm'], stdout=subprocess.PIPE)
        process = subprocess.Popen([
            'sshpass', '-p', 'frdm', 'ssh', '-o',
            'UserKnownHostsFile=/dev/null', '-o', 'StrictHostKeyChecking=no',
            '-t', '-t', '-p',
            str(test_ssh_port), 'fbx@127.0.0.1', 'sudo plinth --list-modules'
        ],
                                   stdin=echo.stdout)
        process.communicate()
    finally:
        echo = subprocess.Popen(['echo', 'frdm'], stdout=subprocess.PIPE)
        process = subprocess.Popen([
            'sshpass', '-p', 'frdm', 'ssh', '-o',
            'UserKnownHostsFile=/dev/null', '-o', 'StrictHostKeyChecking=no',
            '-t', '-t', '-p',
            str(test_ssh_port), 'fbx@127.0.0.1', 'sudo shutdown now'
        ],
                                   stdin=echo.stdout)
        process.communicate()
        time.sleep(30)
        _run(['VBoxManage', 'modifyvm', vm_name, '--hda', 'none'],
             ignore_errors=True)
        _run(['VBoxManage', 'unregistervm', vm_name, '--delete'],
             ignore_errors=True)



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Worker class to build Raspberry Pi 3 Model B+ image.
"""

from .raspberry_pi_3 import RaspberryPi3ImageBuilder


class RaspberryPi3BPlusImageBuilder(RaspberryPi3ImageBuilder):
    """Image builder for Raspberry Pi 3 Model B+ target."""
    machine = 'raspberry3-b-plus'
    flash_kernel_name = 'Raspberry Pi 3 Model B+'



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Base worker class to build all Allwinner A20 based SBC images.
"""

from .. import library
from .arm import ARMImageBuilder


class A20ImageBuilder(ARMImageBuilder):
    """Base image builder for all Allwinner A20 board based targets."""
    architecture: str = 'armhf'
    kernel_flavor: str = 'armmp-lpae'
    u_boot_path: str | None = None
    u_boot_variant: str = 'sunxi'

    def install_boot_loader(self, state):
        """Install the boot loader onto the image."""
        if not self.u_boot_path:
            raise NotImplementedError

        library.install_boot_loader_part(state,
                                         self.u_boot_path,
                                         seek='8',
                                         size='1k')



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Base worker class to run various commands that build the image.
"""

import logging
import os

from . import internal, library

# initramfs-tools is a dependency for the kernel-image package. However, when
# kernel is not installed, as in case of Raspberry Pi image, explicit
# dependency is needed.
BASE_PACKAGES = [
    'initramfs-tools',
]

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class ImageBuilder:  # pylint: disable=too-many-instance-attributes
    """Base for all image builders."""
    architecture: str | None = None
    machine: str = 'all'
    include_non_free_firmware: bool = False
    include_contrib: bool = False

    builder_backend: str = 'internal'
    partition_table_type: str = 'msdos'
    root_filesystem_type: str = 'btrfs'
    boot_filesystem_type: str | None = None
    boot_size: str | None = None
    efi_filesystem_type: str | None = None
    efi_size: str | None = None
    firmware_filesystem_type: str | None = None
    firmware_size: str | None = None
    kernel_flavor: str = 'default'
    debootstrap_variant: str | None = None

    extra_storage_size: str = '1000M'

    @classmethod
    def get_target_name(cls):
        """Return the command line name of target for this builder."""
        return None

    @classmethod
    def get_builder_class(cls, target):
        """Return an builder class given target name."""
        from . import builders  # noqa: F401, pylint: disable=unused-variable

        for subclass in cls.get_subclasses():
            if subclass.get_target_name() == target:
                return subclass

        raise ValueError('No such target')

    @classmethod
    def get_subclasses(cls):
        """Iterate through the subclasses of this class."""
        for subclass in cls.__subclasses__():
            yield subclass
            yield from subclass.get_subclasses()

    def __init__(self, arguments):
        """Initialize object."""
        self.arguments = arguments
        self.packages = BASE_PACKAGES
        self.ram_directory = None

        self.builder_backends = {}
        self.builder_backends['internal'] = internal.InternalBuilderBackend(
            self)

        self.image_file = os.path.join(self.arguments.build_dir,
                                       self._get_image_base_name() + '.img')

    @property
    def image_size(self):
        """Return the size of the image to create based on CLI arguments."""
        return self.arguments.image_size

    @property
    def release_components(self):
        """Return the Debian release components to use for the build."""
        components = ['main']
        if self.include_non_free_firmware:
            components.append('non-free-firmware')

        if self.include_contrib:
            components.append('contrib')

        if self.arguments.release_component:
            for component in self.arguments.release_component:
                if component not in components:
                    components.append(component)

        return components

    def build(self):
        """Run the image building process."""
        archive_file = self.image_file + '.xz'
        self.make_image()
        self.compress(archive_file, self.image_file)

        self.sign(archive_file)

    def _get_builder_backend(self):
        """Returns the builder backend."""
        builder = self.builder_backend
        return self.builder_backends[builder]

    def make_image(self):
        """Call a builder backend to create basic image."""
        self._get_builder_backend().make_image()

    def _get_image_base_name(self):
        """Return the base file name of the final image."""
        build_stamp = self.arguments.build_stamp
        build_stamp = build_stamp + '_' if build_stamp else ''
        return 'freedombox-{distribution}_{build_stamp}{machine}' \
            '-{architecture}'.format(
                distribution=self.arguments.distribution,
                build_stamp=build_stamp, machine=self.machine,
                architecture=self.architecture)

    def compress(self, archive_file, image_file):
        """Compress the generated image."""
        if not self.arguments.skip_compression:
            library.compress(archive_file, image_file)
        else:
            logger.info('Skipping image compression')

    def sign(self, archive):
        """Sign the final output image."""
        if not self.arguments.sign:
            return

        library.sign(archive)

    @staticmethod
    def _replace_extension(file_name, new_extension):
        """Replace a file's extension with a new extention."""
        return file_name.rsplit('.', maxsplit=1)[0] + new_extension
